class RadixTreeNode:
    def __init__(self):
        self.children = {}
        self.host = None
        self.data = None


class BaseRadixTree:
    def __init__(self):
        self.root = RadixTreeNode()



import time
import random
import logging
import ipaddress
from pathlib import Path

log = logging.getLogger("radixtarget.test")

cidr_list_path = Path(__file__).parent / "cidrs.txt"

from radixtarget import RadixTarget


def test_radixtarget():
    rt = RadixTarget()

    for _ in range(2):

        # ipv4
        rt.insert("192.168.1.0/24")
        assert rt.search("192.168.1.10") == ipaddress.ip_network("192.168.1.0/24")
        assert rt.search("192.168.2.10") is None
        rt.insert(ipaddress.ip_network("10.0.0.0/8"))
        assert rt.search("10.255.255.255") == ipaddress.ip_network("10.0.0.0/8")
        rt.insert(ipaddress.ip_network("172.16.12.1"))
        assert rt.search("172.16.12.1") == ipaddress.ip_network("172.16.12.1/32")
        rt.insert("8.8.8.0/24", "custom_data_8")
        assert rt.search("8.8.8.8") == "custom_data_8"

        # ipv6
        rt.insert("dead::/64")
        assert rt.search("dead::beef") == ipaddress.ip_network("dead::/64")
        assert rt.search("dead:cafe::beef") == None
        rt.insert("cafe::babe")
        assert rt.search("cafe::babe") == ipaddress.ip_network("cafe::babe/128")
        rt.insert("beef::/120", "custom_beef")
        assert rt.search("beef::bb") == "custom_beef"

        # networks
        rt.insert("192.168.128.0/24")
        assert rt.search("192.168.128.0/28") == ipaddress.ip_network("192.168.128.0/24")
        assert rt.search("192.168.128.0/23") == None
        rt.insert("babe::/64")
        assert rt.search("babe::/96") == ipaddress.ip_network("babe::/64")
        assert rt.search("babe::/63") == None

        # ipv4 / ipv6 confusion
        rand_int = random.randint(0, 2**32 - 1)
        ipv4_address = ipaddress.IPv4Address(rand_int)
        ipv6_address = ipaddress.IPv6Address(rand_int << (128 - 32))
        ipv6_network = ipaddress.IPv6Network(f"{ipv6_address}/32")
        rt.insert(ipv4_address)
        assert rt.search(ipv4_address)
        assert not rt.search(ipv6_address)
        assert not rt.search(ipv6_network)

        # dns
        rt.insert("net")
        rt.insert("www.example.com")
        rt.insert("test.www.example.com")
        assert rt.search("net") == "net"
        assert rt.search("evilcorp.net") == "net"
        assert rt.search("www.example.com") == "www.example.com"
        assert rt.search("asdf.test.www.example.com") == "test.www.example.com"
        assert rt.search("example.com") is None
        rt.insert("evilcorp.co.uk", "custom_data")
        assert rt.search("www.evilcorp.co.uk") == "custom_data"

        # speed benchmark
        cidrs = open(cidr_list_path).read().splitlines()
        log.critical(len(cidrs))
        for c in cidrs:
            rt.insert(c)

        iterations = 10000

        start = time.time()
        for i in range(iterations):
            random_ip = ipaddress.ip_address(random.randint(0, 2**32 - 1))
            rt.search(random_ip)
        end = time.time()
        elapsed = end - start
        log.critical(
            f"{iterations:,} iterations in {elapsed:.4f} seconds ({int(iterations/elapsed)}/s)"
        )



from .base import BaseRadixTree, RadixTreeNode


class DNSRadixTree(BaseRadixTree):
    def insert(self, hostname, data=None):
        if data is None:
            data = hostname
        parts = hostname.split(".")
        node = self.root
        # Reverse the parts of the hostname for proper hierarchy (TLD to subdomain)
        for part in reversed(parts):
            if part not in node.children:
                node.children[part] = RadixTreeNode()
            node = node.children[part]
        node.data = data

    def search(self, hostname):
        parts = hostname.split(".")
        node = self.root
        matched_data = None
        # Search through the tree in the order from TLD to subdomain
        for part in reversed(parts):
            if part in node.children:
                node = node.children[part]
                if node.data:
                    matched_data = node.data
            else:
                break
        return matched_data



import ipaddress

from .base import BaseRadixTree, RadixTreeNode


class IPRadixTree(BaseRadixTree):

    def insert(self, network, data=None):
        network = ipaddress.ip_network(network, strict=False)
        if data is None:
            data = network
        node = self.root
        network_value = int(network.network_address)
        for i in range(network.prefixlen):
            current_bit = (network_value >> (network.max_prefixlen - 1 - i)) & 1
            if current_bit not in node.children:
                node.children[current_bit] = RadixTreeNode()
            node = node.children[current_bit]
        node.host = network
        node.data = data

    def search(self, query):
        query_network = ipaddress.ip_network(query, strict=False)
        ip_value = int(query_network.network_address)
        query_prefixlen = query_network.prefixlen

        node = self.root
        matched_data = None
        for i in range(query_prefixlen):
            current_bit = (ip_value >> (query_network.max_prefixlen - 1 - i)) & 1
            if current_bit in node.children:
                node = node.children[current_bit]
                if node.host and node.host.prefixlen <= query_prefixlen:
                    if query_network.network_address in node.host:
                        matched_data = node.data
            else:
                break
        return matched_data



import ipaddress

from .tree.ip import IPRadixTree
from .tree.dns import DNSRadixTree


class RadixTarget:
    def __init__(self):
        self.ip_tree = IPRadixTree()
        self.dns_tree = DNSRadixTree()

    def insert(self, host, data=None):
        host = self.make_ip(host)
        if self.is_ip(host):
            self.ip_tree.insert(host, data=data)
        else:
            self.dns_tree.insert(host, data=data)

    def search(self, host):
        host = self.make_ip(host)
        if self.is_ip(host):
            return self.ip_tree.search(host)
        else:
            return self.dns_tree.search(host)

    def make_ip(self, host):
        try:
            return ipaddress.ip_network(host)
        except Exception:
            return str(host)

    def is_ip(self, host):
        version = getattr(host, "version", 0)
        return version == 4 or version == 6
